/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                         V9938.c                         **/
/**                                                         **/
/** This file contains implementation for the V9938 special **/
/** graphical operations.                                   **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994-1999                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/**                                                         **/
/** Patched for more accurate VDP behaviour by Alex Wulms   **/
/**  12 September 1999                                      **/
/*************************************************************/

#include "V9938.h"
#include <string.h>

static struct
{
  byte *P,Op,Mask,PPB,Shift[4];
  int X,NX,NY,CX,CNX,TX,TY;
} MMC;

static byte VDPOp(register byte A,register byte B,register byte N);
static void VDPCopyF(register byte *D,register byte *S,register int N);
static void VDPCopyB(register byte *D,register byte *S,register int N);

/** VDPOp() **************************************************/
/** This function is used by VDPDraw() for operations on    **/
/** pixels.                                                 **/
/*************************************************************/
byte VDPOp(byte A,byte B,byte N)
{
  /* Choose an operation to perform */
  switch(N)
  {
    case 0:   return(A);
    case 1:   return(A&B);
    case 2:   return(A|B);
    case 3:   return(A^B);
    case 4:   return(~A);
    case 8:   return(A? A:B);
    case 9:   return(A? A&B:B);
    case 10:  return(A? A|B:B);
    case 11:  return(A? A^B:B);
    case 12:  return(A? ~A:B);
  }

  /* By default, do a NOP */
  return(B);
}

/** VDPCopyB()/VDPCopyF() ************************************/
/** These two functions called by VDPDraw() copy an array   **/
/** of N bytes from S to D in different directions.         **/
/*************************************************************/
void VDPCopyF(byte *D,byte *S,int N) { for(;N>0;N--) *D++=*S++; }
void VDPCopyB(byte *D,byte *S,int N) { for(;N>0;N--) *D--=*S--; }

/** VDPWrite() ***********************************************/
/** Use this function to transfer pixel(s) from CPU to VDP. **/
/*************************************************************/
void VDPWrite(byte V)
{
  register byte *P;
  register int J;

  /* If no transfer, return */
  if(!(VDPStatus[2]&0x80)) return;

  switch(MMC.Op&0xF0)
  {
    case 0xB0: /* TSFR */
      V&=MMC.Mask;
      if(V||!(MMC.Op&0x08))
      {
        J=MMC.Shift[MMC.CX%MMC.PPB];
        P=MMC.P+MMC.CX/MMC.PPB;
        *P=(*P&~(MMC.Mask<<J))|VDPOp(V<<J,*P,MMC.Op&0x07);
      }
      MMC.CX+=MMC.TX;
      MMC.CNX--;
      break;

    case 0xF0: /* FTSFR */
      MMC.P[MMC.CX]=V;
      MMC.CX+=MMC.TX;
      MMC.CNX--;
      break;

    default:
      return;
  }

  /* If we finished with a line... */
  if(MMC.CNX<=0)
    if(!--MMC.NY) VDPStatus[2]&=0x7E;
    else
    {
      /* Go to the next line */
      MMC.P  += MMC.TY;
      MMC.CX  = MMC.X;
      MMC.CNX = MMC.NX;
    }
}

/** VDPRead() ************************************************/
/** Use this function to transfer pixel(s) from VDP to CPU. **/
/*************************************************************/
byte VDPRead(void)
{
  register byte J;

  /* If no transfer, return 00h */
  if(!(VDPStatus[2]&0x80)) return(0x00);

  switch(MMC.Op&0xF0)
  {
    case 0xA0: /* READ */
      J=(MMC.P[MMC.CX/MMC.PPB]>>MMC.Shift[MMC.CX%MMC.PPB])&MMC.Mask;
      MMC.CX+=MMC.TX;
      MMC.CNX--;
      break;

    default:
      return(0x00);
  }

  /* If we finished with a line... */
  if(MMC.CNX<=0)
    if(!--MMC.NY) VDPStatus[2]&=0x7E;
    else
    {
      /* Go to the next line */
      MMC.P  += MMC.TY;
      MMC.CX  = MMC.X;
      MMC.CNX = MMC.NX;
    }

  /* Return read pixel */
  return(J);
}

/** VDPDraw() ************************************************/
/** Perform a given V9938 operation Op.                     **/
/*************************************************************/
byte VDPDraw(byte Op)
{
  static char *Ops[16] =
  {
    "SET ","AND ","OR  ","XOR ","NOT ","NOP ","NOP ","NOP ",
    "SETT","ANDT","ORT ","XORT","NOTT","NOP ","NOP ","NOP "
  };
  static char *Commands[16] =
  {
    " ABRT"," ????"," ????"," ????"," PGET"," PSET"," SRCH"," LINE",
    " FILL"," COPY"," READ"," TSFR","FFILL","FCOPY","HCOPY","FTSFR"
  };

  static byte Shift[4][4] = { {4,0,0,0},{6,4,2,0},{4,0,0,0},{0,0,0,0} };
  static byte Mask[4]     = { 0x0F,0x03,0x0F,0xFF };
  static int  LPS[4]      = { 1024,1024,512,512 };
  static int  PPB[4]      = { 2,4,2,1 };
  static int  PPL[4]      = { 256,512,512,256 };
  static int  BPL[4]      = { 128,128,256,256 };

  register int AX,AY,TX,TY,I,J,K;
  register int SX,SY,DX,DY,NX,NY,SM,TR;
  register byte *P,*T,CL;

  /* V9938 ops only work in SCREENs 5-8 */
  if(ScrMode<5) return(0);

  SM = ScrMode-5;         /* Screen mode index 0..3  */
  CL = VDP[44]&Mask[SM];  /* Color                   */
  TR = !(Op&0x08);        /* Transparency flag       */

  /* Fetch arguments */
  SX = (VDP[32]+((int)VDP[33]<<8)) & 511;
  SY = (VDP[34]+((int)VDP[35]<<8))%LPS[SM];
  DX = (VDP[36]+((int)VDP[37]<<8)) & 511;
  DY = (VDP[38]+((int)VDP[39]<<8))%LPS[SM];
  NX = (VDP[40]+((int)VDP[41]<<8))%1024;
  NY = (VDP[42]+((int)VDP[43]<<8))%1024;

  if(Verbose&0x02)
    printf
    (
      "V9938: Opcode %02Xh %s-%s (%d,%d)->(%d,%d),%d [%d,%d]%s\n",
      Op,Commands[Op>>4],Ops[Op&0x0F],SX,SY,DX,DY,CL,
      VDP[45]&0x04? -NX:NX,VDP[45]&0x08? -NY:NY,
      VDP[46]&0x20? " on ExtVRAM":""
    );

  switch(Op&0xF0)
  {

/** ABRT *****************************************************/
/** Abort current V9938 operation.                          **/
/*************************************************************/
case 0x00:
  VDPStatus[2]&=0x7E;
  break;

/** PGET *****************************************************/
/** Get color of a point at SX,SY.                          **/
/*************************************************************/
case 0x40:
  /* Map SX in range */
  SX &= (PPL[SM]-1);
 
  /* Compute VRAM address */
  P=VRAM+SY*BPL[SM]+SX/PPB[SM];
  /* Get the point */
  VDPStatus[7]=VDP[44]=(*P>>Shift[SM][SX%PPB[SM]])&Mask[SM];
  /* Done with PGET */
  break;

/** PSET *****************************************************/
/** Set color of a point at DX,DY.                          **/
/*************************************************************/
case 0x50:
  /* Don't set transparent color */
  if(!CL&&!TR) break;

  /* Map DX in range */
  DX &= (PPL[SM]-1);
 
  /* Compute VRAM address */
  P=VRAM+DY*BPL[SM]+DX/PPB[SM];
  /* Compute point offset */
  I=Shift[SM][DX%PPB[SM]];
  /* Set the point */
  *P=(*P&~(Mask[SM]<<I))|VDPOp(CL<<I,*P,Op&0x07);
  /* Done with PSET */
  break;

/** SRCH *****************************************************/
/** Search from SX,SY in a given direction for a point of a **/
/** given color.                                            **/
/*************************************************************/
case 0x60:
  /* Compute line address in VRAM */
  P=VRAM+SY*BPL[SM];

  /* assume not found */
  VDPStatus[2]&=0xEF;

  /* Do we look for "==" or "!="? */
  I=(VDP[45]&0x02)!=0;

  /* Searching... */
  TX=(VDP[45]&0x04) ? -1:1;
  AX=PPL[SM];
  do {
    J=(P[(SX&(AX-1))/PPB[SM]]>>Shift[SM][SX%PPB[SM]])&Mask[SM];
    if((J==CL)^I) {
      VDPStatus[2]|=0x10;
      break;
    }
    SX += TX;
  } while((SX & AX)!=AX);

  /* Update SX in VDP registers */
  VDPStatus[8]=VDP[32]=SX&0xFF;
  VDPStatus[9]=VDP[33]=(SX>>8)|0xFE;
 
  /* Done with SRCH */
  break;

/** LINE *****************************************************/
/** Draw a line of dimensions NX,NY from DX,DY.             **/
/*************************************************************/
case 0x70:
  /* Don't set transparent color */
  if(!CL&&!TR) break;

  /* Compute operation code */
  Op&=0x07;

  /* Mask major/minor direction to 10 bits range */
  NX&=1023;
  NY&=1023;

  /* Prepare to draw a line */
  J=((NX-1)>>1);
  TX=VDP[45]&0x04?-1:1;
  TY=VDP[45]&0x08?-1:1;
  AX=PPL[SM];
  AY=LPS[SM]-1;

  /* Draw the line */
  if ((VDP[45]&0x01)==0) {
    /* X-Axis is major direction */
    K=0;
    do {
      /* Set a point */
      P=VRAM+(DY&AY)*BPL[SM]+(DX%PPL[SM])/PPB[SM];
      I=Shift[SM][DX%PPB[SM]];
      *P=(*P&~(Mask[SM]<<I))|VDPOp(CL<<I,*P,Op);
      /* Determine next coordinate */
      DX+=TX;
      if ((J-=NY)<0) {
        J+=NX;
        DY+=TY;
      }
      J&=1023; /* Mask to 10 bits range */
    } while ((K++!=NX) && ((DX & AX)!=AX));
  }
  else {
    /* Y-Axis is major direction */
    K=0;
    do {
      /* Set a point */
      P=VRAM+(DY&AY)*BPL[SM]+(DX%PPL[SM])/PPB[SM];
      I=Shift[SM][DX%PPB[SM]];
      *P=(*P&~(Mask[SM]<<I))|VDPOp(CL<<I,*P,Op);
      /* Determine next coordinate */
      DY+=TY;
      if ((J-=NY)<0) {
        J+=NX;
        DX+=TX;
      }
      J&=1023; /* Mask to 10 bits range */
    } while ((K++!=NX) && ((DX & AX)!=AX));
  }
 
  /* Done with LINE */
  break;                

/** FFILL ****************************************************/
/** Quickly fill a box with all bits given in the color     **/
/** register. Logical operation is ignored in this case.    **/
/*************************************************************/
case 0xC0:
  /* This operation requires all color bits */
  CL=VDP[44];

  /* Map NX in range */
  NX &= 511;

  /* Map NY in range */
  NY &= 1023;
  if (NY > LPS[SM])
    NY = LPS[SM];
  /* For NY=0, assume entire height */
  if(!NY)
    NY=LPS[SM];

  /* Calculate Y direction */
  TY=VDP[45]&0x08? -1:1;

  /* Convert x-parameters into byte parameters */
  DX /= PPB[SM];
  NX /= PPB[SM];

  if (DX >= BPL[SM]) {
    /* Special cases, DX in range [256...511] */
    /* while legal range is [0...255]         */
    /* Handle it with the routine that matches*/
    /* the real VDP's behaviour most closely  */

    /* Calcylate X Direction */
    TX=VDP[45]&0x04? -1:1;

    /* X-boundary detection mask */
    K=BPL[SM];

    /* Filling... */
    for(;NY && DY!=-1;NY--) {
      DY &= LPS[SM]-1;
      P = VRAM + DY*BPL[SM];
      AX=DX;
      J=NX;
      do {
        *(P + (AX & BPL[SM]-1)) = CL;
        AX += TX;
        J--;
      } while ((J != 0) && ((AX & K)!=K));
      DY += TY;
    }
  }
  else {
    /* Normal cases, DX in legal range */
    /* Use memset() function to fill   */
    /* up the individual lines         */

    /* For NX=0, assume fill until border */
    if(!NX) { NX=VDP[45]&0x04? DX+1:BPL[SM]-DX; }

    /* Make sure to fill-up from left to right */
    DX=VDP[45]&0x04? DX-NX+1:DX;
    if (DX<0) {
      NX+=DX;
      DX=0;
    }
    /* Map NX in range */
    NX=DX+NX>BPL[SM]? BPL[SM]-DX:NX;

    /* Filling... */
    P = VRAM + DX;
    for(;NY && DY!=-1;NY--) {
      memset(P + (DY & (LPS[SM]-1))*BPL[SM],CL,NX);
      DY += TY;
    }
  }
  /* Done with FFILL */
  break;

/** FILL *****************************************************/
/** Fill a box with a given color.                          **/
/*************************************************************/
case 0x80:
  /* Don't set transparent color */
  if(!CL&&!TR) break;

  /* Compute operation code */
  Op&=0x07;

  /* Map NX in range */
  NX &= 511;

  /* Map NY in range */
  NY &= 1023;
  /* For NY=0, assume entire height */
  if(!NY)
    NY=1024;

  /* Calculate X and Y direction */
  TX=VDP[45]&0x04? -1:1;
  TY=VDP[45]&0x08? -1:1;

  /* X-boundary detection mask */
  K=PPL[SM];

  /* Filling... */
  for(;NY && DY!=-1;NY--) {
    DY &= LPS[SM]-1;
    T = VRAM + DY*BPL[SM];
    AX=DX;
    J=NX;
    do {
      P=T+(AX & (PPL[SM]-1))/PPB[SM];
      I=Shift[SM][AX%PPB[SM]];
      *P=(*P&~(Mask[SM]<<I))|VDPOp(CL<<I,*P,Op);
      AX += TX;
      J--;
    } while ((J != 0) && ((AX & K)!=K));
    DY += TY;
  }
 
  /* Done with FILL */
  break;

/** HCOPY ****************************************************/
/** Quickly copy a block of pixels at coinciding horizontal **/
/** coordinate SX=DX. Logical operation is ignored in this  **/
/** case.                                                   **/
/*************************************************************/
case 0xE0:
  /* Map NY in range */
  NY &= 1023;
  /* For NY=0, assume entire height */
  if(!NY)
    NY=1024;

  /* Calculate X Direction */
  TX=VDP[45]&0x04? -1:1;

  /* Calculate Y direction */
  TY=VDP[45]&0x08? -1:1;

  /* Convert x-parameters into byte parameters */
  DX /= PPB[SM];

  if (DX >= BPL[SM]) {
    /* Special cases, DX in range [256...511] */
    /* while legal range is [0...255]         */
    /* Handle it with the routine that matches*/
    /* the real VDP's behaviour most closely  */

    /* X-boundary detection mask */
    K=BPL[SM];

    /* Copying... */
    for(;NY && SY != -1 && DY!=-1;NY--) {
      SY &= LPS[SM]-1;
      DY &= LPS[SM]-1;
      T = VRAM + SY*BPL[SM];
      P = VRAM + DY*BPL[SM];
      AX=DX;
      do {
        *(P + (AX & BPL[SM]-1)) = *(T + (AX & BPL[SM]-1));
        AX += TX;
      } while ((AX & K)!=K);
      SY += TY;
      DY += TY;
    }
  }
  else {
    /* Normal case, DX in legal range */
    /* Use VdpCopy() function to copy */
    /* the individual lines           */

    /* Copy until border */
    NX=VDP[45]&0x04? DX+1:BPL[SM]-DX;

    /* Based on horizontal direction, set horizontal step */
    TX=VDP[45]&0x04? -1:1;

    /* Copying... */
    for(;NY && SY != -1 && DY!=-1;NY--) {
      SY &= LPS[SM]-1;
      DY &= LPS[SM]-1;
      T = VRAM + SY*BPL[SM];
      P = VRAM + DY*BPL[SM];
      if (TX>0)
        VDPCopyF(P+DX, T+DX, NX);
      else
        VDPCopyB(P+DX, T+DX, NX);
      SY += TY;
      DY += TY;
    }
  }
 
  /* Done with HCOPY */
  break;

/** COPY *****************************************************/
/** Copy a NXxNY block of pixels from SX,SY to DX,DY in a   **/
/** given directions.                                       **/
/*************************************************************/
case 0x90:
  /* Map SX in range */
  SX &= (PPL[SM]-1);
  DX &= (PPL[SM]-1);
 
  /* Compute operation code */
  Op&=0x07;

  /* For null dimensions, assume entire width/height */
  if(!NX) { NX=PPL[SM];DX=VDP[45]&0x04? NX-1:0;SX=VDP[45]&0x04? NX-1:0; }
  if(!NY) { NY=LPS[SM];DY=VDP[45]&0x08? NY-1:0;SY=VDP[45]&0x08? NY-1:0; }

  /* Compute source and destination pointers */
  T=VRAM+SY*BPL[SM];
  P=VRAM+DY*BPL[SM];

  /* Based on horizontal direction... */
  if(VDP[45]&0x04)
  {
    /* Set horizontal step */
    TX=-1;
    /* Truncate block width */
    if(SX-NX<-1) NX=SX;
    if(DX-NX<-1) NX=DX;
  }
  else
  {
    /* Set horizontal step */
    TX=1;
    /* Truncate block width */
    if(SX+NX>PPL[SM]) NX=PPL[SM]-SX;
    if(DX+NX>PPL[SM]) NX=PPL[SM]-DX;
  }

  /* Based on vertical direction... */
  if(VDP[45]&0x08)
  {
    /* Set vertical step */
    TY=-BPL[SM];
    /* Truncate block height */
    if(DY-NY<-1) NY=DY+1;
    if(SY-NY<-1) NY=SY+1;
  }
  else
  {
    /* Set vertical step */
    TY=BPL[SM];
    /* Truncate block height */
    if(DY+NY>LPS[SM]) NY=LPS[SM]-DY;
    if(SY+NY>LPS[SM]) NY=LPS[SM]-SY;
  }

  /* Copying... */
  for(;NY;NY--,P+=TY,T+=TY)
    for(J=0;J<NX;J++)
    {
      /* Get source point */
      AX=SX+J*TX;
      CL=(T[AX/PPB[SM]]>>Shift[SM][AX%PPB[SM]])&Mask[SM];

      /* If drawing this point... */
      if(CL||TR)
      {
        AX=DX+J*TX;
        K=AX/PPB[SM];
        I=Shift[SM][AX%PPB[SM]];
        P[K]=(P[K]&~(Mask[SM]<<I))|VDPOp(CL<<I,P[K],Op);
      }
    }

  /* Done with COPY */
  break;

/** FCOPY ****************************************************/
/** Copy a NXxNY block of pixels from SX,SY to DX,DY in a   **/
/** given directions. Logical operation is ignored in this  **/
/** case.                                                   **/
/*************************************************************/
case 0xD0:
  /* Map SX in range */
  SX &= (PPL[SM]-1);
  DX &= (PPL[SM]-1);
 
  /* Set SX,DX,NX to number of bytes */
  SX/=PPB[SM];
  DX/=PPB[SM];
  NX/=PPB[SM];

  /* For null dimensions, assume entire width/height */
  if(!NX) { NX=BPL[SM];SX=DX=VDP[45]&0x04? BPL[SM]-1:0; }
  if(!NY) { NY=LPS[SM];SY=DY=VDP[45]&0x08? NY-1:0; }

  /* Based on horizontal direction... */
  if(VDP[45]&0x04)
  {
    /* Set horizontal step */
    TX=-1;
    /* Truncate block width */
    if(SX-NX<-1) NX=SX;
    if(DX-NX<-1) NX=DX;
  }
  else
  {
    /* Set horizontal step */
    TX=1;
    /* Truncate block width */
    if(SX+NX>BPL[SM]) NX=BPL[SM]-SX;
    if(DX+NX>BPL[SM]) NX=BPL[SM]-DX;
  }

  /* Based on vertical direction... */
  if(VDP[45]&0x08)
  {
    /* Set vertical step */
    TY=-BPL[SM];
    /* Truncate block height */
    if(DY-NY<-1) NY=DY+1;
    if(SY-NY<-1) NY=SY+1;
  }
  else
  {
    /* Set vertical step */
    TY=BPL[SM];
    /* Truncate block height */
    if(DY+NY>LPS[SM]) NY=LPS[SM]-DY;
    if(SY+NY>LPS[SM]) NY=LPS[SM]-SY;
  }

  /* Compute source and destination pointers */
  T=VRAM+SY*BPL[SM]+SX;
  P=VRAM+DY*BPL[SM]+DX;

  /* Copying... */
  for(;NY;NY--,P+=TY,T+=TY)
    if(TX>0) VDPCopyF(P,T,NX); else VDPCopyB(P,T,NX);

  /* Done with FCOPY */
  break;

/** TSFR *****************************************************/
/** Transfer a block of pixels to/from CPU.                 **/
/*************************************************************/
case 0xA0:
case 0xB0:
case 0xF0:
  /* Map SX in range */
  SX &= (PPL[SM]-1);
  DX &= (PPL[SM]-1);
 
  /* Extract command part of the opcode */
  J=Op&0xF0;

  /* For READ, we use SX,SY */
  if(J==0xA0) { DX=SX;DY=SY; }

  /* For null dimensions, assume entire width/height */
  if(!NX) { NX=PPL[SM];DX=VDP[45]&0x04? NX-1:0; }
  if(!NY) { NY=LPS[SM];DY=VDP[45]&0x08? NY-1:0; }

  /* Based on horizontal direction... */
  if(VDP[45]&0x04)
  {
    /* Set horizontal step */
    TX=-1;
    /* Truncate block width */
    if(DX-NX<-1) NX=DX;
  }
  else
  {
    /* Set horizontal step */
    TX=1;
    /* Truncate block width */
    if(DX+NX>PPL[SM]) NX=PPL[SM]-DX;
  }

  /* Based on vertical direction... */
  if(VDP[45]&0x08)
  {
    /* Set vertical step */
    TY=-BPL[SM];
    /* Truncate block height */
    if(DY-NY<-1) NY=DY+1;
  }
  else
  {
    /* Set vertical step */
    TY=BPL[SM];
    /* Truncate block height */
    if(DY+NY>LPS[SM]) NY=LPS[SM]-DY;
  }

  if(J==0xF0)
  {
    /* Compute byte values for the byte-wise transfers */
    DX/=PPB[SM];NX/=PPB[SM];
  }
  else
  {
    /* Copy mask/shift values for pixel-wise transfers */
    memcpy(MMC.Shift,Shift[SM],4);
    MMC.PPB  = PPB[SM];
    MMC.Mask = Mask[SM];
  }

  /* Assign values to the transfer variables */
  MMC.Op   = Op;
  MMC.P    = VRAM+DY*BPL[SM];
  MMC.X    = DX;
  MMC.NX   = NX;
  MMC.NY   = NY;
  MMC.CX   = MMC.X;
  MMC.CNX  = MMC.NX;
  MMC.TX   = TX;
  MMC.TY   = TY;

  /* Start VDP operation */
  VDPStatus[2]|=0x81;

  /* Transfer first byte */
  if(J!=0xA0) VDPWrite(VDP[44]);

  /* Done with TSFR */
  break;

/** UNKNOWN **************************************************/
/** Unrecognized V9938 operation.                           **/
/*************************************************************/
default:
  if(Verbose&0x02) printf("V9938: Unrecognized opcode %02Xh\n",Op);
  return(0);

  }

  /* Operation successfull */
  return(1);
}
